/*
 * Rahmenwerk-Plug-in "Konfigurationsassistent"
 *
 * Copyright (C) 2018 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weissenfelser Strasse 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 *
 * Copyright (c) 2011 SWARCO TRAFFIC SYSTEMS GmbH.
 * 
 * Contributors:
 * SWARCO TRAFFIC SYSTEMS GmbH - initial API and implementation
 *
 * Contact Information:
 * SWARCO TRAFFIC SYSTEMS GmbH
 * Adolf-Dambach-Straße
 * D-76571 Gaggenau
 * Phone: +49-7225-64-03
 * Fax: +49-7225-64-300
 * mailto: office.sts@swarco.de
 */

package de.bsvrz.buv.plugin.konfigass;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;

import de.bsvrz.dav.daf.main.config.ConfigurationArea;
import de.bsvrz.dav.daf.main.config.ConfigurationChangeException;
import de.bsvrz.dav.daf.main.config.ConfigurationTaskException;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.management.ConfigAreaAndVersion;
import de.bsvrz.dav.daf.main.config.management.ConfigurationControl;
import de.bsvrz.dav.daf.main.config.management.consistenycheck.ConsistencyCheckResultInterface;
import de.bsvrz.puk.config.configFile.datamodel.AreaDependencyChecker;
import de.bsvrz.puk.config.configFile.datamodel.ConfigConfigurationArea;
import de.bsvrz.puk.config.configFile.datamodel.ConfigDataModel;

/**
 * Mit dieser Klasse wird die Konfiguration gestartet. Die Konfiguration kann in
 * folgenden verschiedenen Arbeitsmodi parallel gestartet werden:
 * <ul>
 * <li>Import von Versorgungsdateien</li>
 * <li>Export von Versorgungsdateien</li>
 * <li>Konsistenzprüfung der Konfiguration</li>
 * <li>Aktivierung von Konfigurationsbereichen</li>
 * <li>Freigabe zur Übernahme von Konfigurationsbereichen</li>
 * <li>Freigabe zur Aktivierung von Konfigurationsbereichen</li>
 * <li>Konfigurationseditor zur interaktiven Versorgung</li>
 * <li>Konfiguration als dienstleistender Hintergrundprozeß für andere
 * Applikationen des Systems</li>
 * </ul>
 * <p/>
 * Die Konfiguration kann mit folgenden spezifischen Aufrufparametern gestartet
 * werden:
 * <ul>
 * <li>-verwaltung=datei</li>Über diesen Aufrufparameter wird die
 * Verwaltungsdatei spezifiziert, aus der sich die Konfiguration initialisieren
 * soll. Wenn keine weiteren Aufrufparameter angegeben werden, dann wird der
 * Arbeitsmodus 'Konfiguration' gestartet.
 * <ul>
 * Soll Die Konfiguration im Arbeitsmodus 'Konfiguration' betrieben werden, so
 * können/müssen folgende Parameter angegeben werden:<br>
 * <li>Optionaler Parameter: -benutzerverwaltung=datei</li>Über diesen Parameter
 * wird die Datei spezifiziert, die die
 * Benutzerverwaltungsinformationen(Benutzernamen, Passwörter, usw.) enthält.
 * Wird keine Datei angegeben, wird eine Default-Datei benutzt. Neue Benutzer,
 * Änderungen der Passwörter, usw. werden in der angegebenen Datei durchgeführt.
 * <br>
 * <li>-benutzer=Benutzername</li> Enthält den Benutzernamen, mit dem sich die
 * Konfiguration beim Datenverteiler anmeldet
 * <li>-authentifizierung=datei</li> Enthält das Passwort, mit dem sich die
 * Konfiguration beim Datenverteiler anmeldet
 * </ul>
 * <li>-import=pid,pid,...</li>Über diesen Aufrufparameter werden eine oder
 * mehrere Pids der Konfigurationsbereiche angegeben, die importiert werden
 * sollen. Wenn dieser Aufrufparameter angegeben wurde, dann wird der
 * Arbeitsmodus 'Import' gestartet. In diesem Arbeitsmodus müssen außerdem die
 * Aufrufparameter '-verzeichnis' und '-verwaltung' angegeben werden.
 * <li>-export=pid,pid,...</li>Über diesen Aufrufparameter werden eine oder
 * mehrere Pids der Konfigurationsbereiche angegeben, die exportiert werden
 * sollen. Wenn dieser Aufrufparameter angegeben wurde, dann wird der
 * Arbeitsmodus 'Export' gestartet. In diesem Arbeitsmodus müssen außerdem die
 * Aufrufparameter '-verzeichnis' und '-verwaltung' angegeben werden.
 * <li>-verzeichnis=verzeichnis</li>Dieser Aufrufparameter gilt im Zusammenhang
 * mit dem Arbeitsmodus 'Import' oder 'Export'. Er gibt an, wo (beim Import)
 * sich die Dateien befinden bzw. wohin (beim Export) die Dateien geschrieben
 * werden sollen.
 * <li>-editor</li>Über diesen Aufrufparameter wird spezifiziert, dass der
 * Arbeitsmodus 'Konfigurationseditor' gestartet werden soll. Wenn zusätzlich
 * der Aufrufparameter '-verwaltung' angegeben wird, dann arbeitet der Editor im
 * Offline-Modus direkt auf den Konfigurationsdateien. Ansonsten arbeitet der
 * Editor im Online-Modus und versucht über die Datenverteilerschnittstelle eine
 * Verbindung zu einer Instanz im Arbeitsmodus 'Konfiguration' aufzubauen.
 * <li>-konsistenzprüfung</li>Beauftragt die Konfiguration alle
 * Konfigurationsbereiche in der aktuellsten Version auf Konsistenz zu prüfen.
 * In diesem Arbeitsmodus muss außerdem der Aufrufparameter '-verwaltung'
 * angegeben werden.
 * <li>-aktivierung</li>Beauftragt die Konfiguration alle Konfigurationsbereiche
 * in der aktuellsten Version zu aktivieren. In diesem Arbeitsmodus muss
 * außerdem der Aufrufparameter '-verwaltung' angegeben werden.
 * <li>-freigabeaktivierung</li>Beauftragt die Konfiguration alle
 * Konfigurationsbereiche in der aktuellsten Version für andere Konfigurationen
 * "zur Aktivierung freizugeben". In diesem Arbeitsmodus muss außerdem der
 * Aufrufparameter '-verwaltung' angegeben werden.
 * <li>-freigabeübernahme</li>Beauftragt die Konfiguration alle
 * Konfigurationsbereiche in der aktuellsten Version zur Übernahme freizugeben.
 * In diesem Arbeitsmodus muss außerdem der Aufrufparameter '-verwaltung'
 * angegeben werden.
 * <li>-freigabeZurAktivierungOhneLokaleAktivierungDurchDenKV=pid,pid,...</li>
 * Über diesen Aufrufparameter werden eine oder mehrere Pids von
 * Konfigurationsbereichen angegeben. Diese Bereiche können durch eine andere
 * Konfiguration aktiviert werden, ohne dass der Konfigurationsverantwortliche
 * der Bereiche die angegebene Bereiche aktiviert hat. In diesem Arbeitsmodus
 * muss außerdem der Aufrufparameter '-verwaltung' angegeben werden.
 * <li>keine Aufrufparameter</li>Ohne Aufrufparameter wird der Arbeitsmodus
 * 'Konfigurationseditor' im Online-Modus gestartet.
 * </ul>
 *
 * @author BitCtrl Systems GmbH, Uwe Peuker, Kappich Systemberatung
 *
 */
public class KonfigApplikation {

	private final DataModel dataModel;

	/**
	 * Der Konstruktor wertet die Aufrufargumente aus.
	 *
	 * @param dataModel das Datenmodell der Konfiguration
	 */
	public KonfigApplikation(final DataModel dataModel) {

		if (dataModel != null) {
			this.dataModel = dataModel;
		} else {
			this.dataModel = null;

			final String meldung = "Datenmodell ist null";

			final IllegalArgumentException e = new IllegalArgumentException();
			KonfigurationsAssistent.getDefault().getLog()
					.log(new Status(IStatus.ERROR, KonfigurationsAssistent.PLUGIN_ID, IStatus.ERROR, meldung, e));
		}
	}

	/**
	 * @param importPath
	 * @param pids
	 * @throws ConfigurationChangeException
	 */
	public void startImport(final File importPath, final List<String> pids) throws ConfigurationChangeException {
		// Alle Dateien sind für andere gesperrt worden (lock-Dateien). Diese
		// müssen wieder freigegeben werden, sobald
		// die Aktion beendet wurde.

		((ConfigurationControl) dataModel).importConfigurationAreas(importPath, pids);
	}

	/**
	 * @param exportPath
	 * @param exportPids
	 * @throws ConfigurationTaskException
	 */
	public void startExport(final File exportPath, final List<String> exportPids) throws ConfigurationTaskException {
		// Alle Dateien sind für andere gesperrt worden (lock-Dateien). Diese
		// müssen wieder freigegeben werden, sobald
		// die Aktion beendet wurde.
		// starteAutomatischesBeenden();

		List<String> pids = new ArrayList<>();
		pids.addAll(exportPids);

		if ((pids.size() == 1) && pids.get(0).isEmpty()) {
			pids = new ArrayList<>(((ConfigDataModel) dataModel).getAllConfigurationAreas().keySet());
		}

		((ConfigDataModel) dataModel).exportConfigurationAreas(exportPath, pids);
	}

	/**
	 * Unterzieht alle Bereiche in der aktuellsten Version einer Konsistenzprüfung.
	 * Die Version muss nicht unbedingt die aktuelle Version sein, sondern die
	 * Version, die aktiviert werden könnte.
	 *
	 * @param allowDoublePids                          <code>true</code> falls
	 *                                                 doppelte Pids in
	 *                                                 verschiedenen
	 *                                                 Konfigurationsbereichen von
	 *                                                 der Konsistenzprüfung
	 *                                                 zugelassen werden sollen.
	 * @param ignoreDependencyErrorsInConsistencyCheck <code>true</code> falls
	 *                                                 Fehler bei der Prüfung der
	 *                                                 Abhängigkeiten in der
	 *                                                 Konsistenzprüfung ignoriert
	 *                                                 werden sollen.
	 * @return das Ergebnis der Konsistenzprüfung
	 */
	public KonfigAssistentErgebnis pruefeKonsistenz(final boolean allowDoublePids,
			final boolean ignoreDependencyErrorsInConsistencyCheck) throws Exception {
		((ConfigDataModel) dataModel).setAllowDoublePids(allowDoublePids);
		((ConfigDataModel) dataModel)
				.setIgnoreDependencyErrorsInConsistencyCheck(ignoreDependencyErrorsInConsistencyCheck);
		// Alle Dateien sind für andere gesperrt worden (lock-Dateien). Diese
		// müssen wieder freigegeben werden, sobald

		final Collection<ConfigurationArea> configurationAreas = ((ConfigDataModel) dataModel)
				.getAllConfigurationAreas().values();
		// Alle Bereiche mit den jeweiligen Versionen
		final List<ConfigAreaAndVersion> configAreasAndVersions = new ArrayList<>();

		for (final ConfigurationArea configurationArea : configurationAreas) {
			final short version;
			if (dataModel.getConfigurationAuthority() == configurationArea.getConfigurationAuthority()) {
				// In der neusten Version prüfen
				version = ((ConfigConfigurationArea) configurationArea).getLastModifiedVersion();
			} else {
				version = configurationArea.getTransferableVersion();
			}
			configAreasAndVersions.add(new ConfigAreaAndVersion(configurationArea, version));
		}

		KonfigAssistentErgebnis ergebnis = new KonfigAssistentErgebnis();
		ergebnis.setBetrachteteBereiche(configAreasAndVersions);
		ergebnis.setAreaDependencyCheckResult(new AreaDependencyChecker().checkAreas(configAreasAndVersions));
		ergebnis.setConsistencyCheckResult(((ConfigurationControl) dataModel).checkConsistency(configAreasAndVersions));
		return ergebnis;
	}

	/**
	 * Aktiviert alle Bereiche in der höchst möglichen Version.
	 *
	 * @param allowDoublePids                          <code>true</code> falls
	 *                                                 doppelte Pids in
	 *                                                 verschiedenen
	 *                                                 Konfigurationsbereichen von
	 *                                                 der Konsistenzprüfung
	 *                                                 zugelassen werden sollen.
	 * @param ignoreDependencyErrorsInConsistencyCheck <code>true</code> falls
	 *                                                 Fehler bei der Prüfung der
	 *                                                 Abhängigkeiten in der
	 *                                                 Konsistenzprüfung ignoriert
	 *                                                 werden sollen.
	 * @throws ConfigurationChangeException
	 */
	public KonfigAssistentErgebnis starteAktivierung(final Collection<ConfigurationArea> configurationAreas,
			final boolean allowDoublePids, final boolean ignoreDependencyErrorsInConsistencyCheck)
			throws ConfigurationChangeException {
		((ConfigDataModel) dataModel).setAllowDoublePids(allowDoublePids);
		((ConfigDataModel) dataModel)
				.setIgnoreDependencyErrorsInConsistencyCheck(ignoreDependencyErrorsInConsistencyCheck);

		final List<ConfigAreaAndVersion> configAreasAndVersions = new ArrayList<>();

		for (final ConfigurationArea configurationArea : configurationAreas) {
			// Den Bereich mit der aktuellsten Version prüfen und aktivieren.
			configAreasAndVersions
					.add(new ConfigAreaAndVersion(configurationArea, configurationArea.getActivatableVersion()));
		}

		KonfigAssistentErgebnis ergebnis = new KonfigAssistentErgebnis();
		ergebnis.setBetrachteteBereiche(configAreasAndVersions);
		ergebnis.setAreaDependencyCheckResult(new AreaDependencyChecker().checkAreas(configAreasAndVersions));
		ergebnis.setConsistencyCheckResult(
				((ConfigurationControl) dataModel).activateConfigurationAreas(configAreasAndVersions));

		return ergebnis;
	}

	/**
	 * Gibt alle Bereiche zur Aktivierung für andere Konfigurationen in der höchst
	 * möglichen Version frei.
	 *
	 * @param configurationAreas die verwendeten Konfigurationsbereiche
	 * @throws ConfigurationChangeException
	 */
	public KonfigAssistentErgebnis starteFreigabeZurAktivierung(final Collection<ConfigurationArea> configurationAreas)
			throws ConfigurationChangeException {

		final List<ConfigAreaAndVersion> configAreasAndVersions = new ArrayList<>(configurationAreas.size());

		for (final ConfigurationArea configurationArea : configurationAreas) {
			// Den Bereich mit der lokal aktivierten Version für andere
			// freigeben
			configAreasAndVersions
					.add(new ConfigAreaAndVersion(configurationArea, configurationArea.getActivatableVersion()));
		}

		KonfigAssistentErgebnis ergebnis = new KonfigAssistentErgebnis();
		ergebnis.setBetrachteteBereiche(configAreasAndVersions);
		((ConfigurationControl) dataModel).releaseConfigurationAreasForActivation(configAreasAndVersions);

		return ergebnis;

	}

	/**
	 * Gibt alle Bereiche zur Übernahme für andere Konfigurationen in der höchst
	 * möglichen Version frei.
	 *
	 * @param configurationAreas                       die verwendeten
	 *                                                 Konfigurationsbereiche
	 * @param allowDoublePids                          <code>true</code> falls
	 *                                                 doppelte Pids in
	 *                                                 verschiedenen
	 *                                                 Konfigurationsbereichen von
	 *                                                 der Konsistenzprüfung
	 *                                                 zugelassen werden sollen.
	 * @param ignoreDependencyErrorsInConsistencyCheck <code>true</code> falls
	 *                                                 Fehler bei der Prüfung der
	 *                                                 Abhängigkeiten in der
	 *                                                 Konsistenzprüfung ignoriert
	 *                                                 werden sollen.
	 * @throws ConfigurationChangeException
	 */
	public KonfigAssistentErgebnis starteFreigabeZurUebernahme(final Collection<ConfigurationArea> configurationAreas,
			final boolean allowDoublePids, final boolean ignoreDependencyErrorsInConsistencyCheck)
			throws ConfigurationChangeException {
		((ConfigDataModel) dataModel).setAllowDoublePids(allowDoublePids);
		((ConfigDataModel) dataModel)
				.setIgnoreDependencyErrorsInConsistencyCheck(ignoreDependencyErrorsInConsistencyCheck);

		final List<ConfigAreaAndVersion> configAreasAndVersions = new ArrayList<>();

		// Es dürfen nur Bereiche freigegeben werden, für die die Konfiguration
		// auch der KV ist.

		for (final ConfigurationArea configurationArea : configurationAreas) {
			if (dataModel.getConfigurationAuthority() == configurationArea.getConfigurationAuthority()) {
				// Der Bereich wird mit der größt möglichen Version für andere
				// zur Übernahme freigegeben
				configAreasAndVersions
						.add(new ConfigAreaAndVersion(configurationArea, configurationArea.getActivatableVersion()));
			}
		}

		KonfigAssistentErgebnis ergebnis = new KonfigAssistentErgebnis();
		ergebnis.setBetrachteteBereiche(configAreasAndVersions);
		ergebnis.setAreaDependencyCheckResult(new AreaDependencyChecker().checkAreas(configAreasAndVersions));
		ergebnis.setConsistencyCheckResult(
				((ConfigurationControl) dataModel).releaseConfigurationAreasForTransfer(configAreasAndVersions));
		return ergebnis;
	}

	/**
	 * Gibt die angegebenen Bereiche zur Aktivierung für andere Konfigurationen in
	 * der höchst möglichen Version frei. Die entsprechenden Bereiche müssen vorher
	 * nicht lokal aktiviert worden sein.
	 *
	 * @param configurationAreas                       die Liste der
	 *                                                 Konfigurationsbereiche
	 * @param allowDoublePids                          <code>true</code> falls
	 *                                                 doppelte Pids in
	 *                                                 verschiedenen
	 *                                                 Konfigurationsbereichen von
	 *                                                 der Konsistenzprüfung
	 *                                                 zugelassen werden sollen.
	 * @param ignoreDependencyErrorsInConsistencyCheck <code>true</code> falls
	 *                                                 Fehler bei der Prüfung der
	 *                                                 Abhängigkeiten in der
	 *                                                 Konsistenzprüfung ignoriert
	 *                                                 werden sollen.
	 * @throws Exception
	 */
	public KonfigAssistentErgebnis startFreigabeZurAktivierungOhneKVAktivierung(
			final Collection<ConfigurationArea> configurationAreas, final boolean allowDoublePids,
			final boolean ignoreDependencyErrorsInConsistencyCheck) throws Exception {
		((ConfigDataModel) dataModel).setAllowDoublePids(allowDoublePids);
		((ConfigDataModel) dataModel)
				.setIgnoreDependencyErrorsInConsistencyCheck(ignoreDependencyErrorsInConsistencyCheck);

		final List<ConfigAreaAndVersion> configAreasAndVersions = new ArrayList<>();

		// Es dürfen nur Bereiche freigegeben werden, für die die Konfiguration
		// auch der KV ist.

		for (final ConfigurationArea configurationArea : configurationAreas) {
			if (dataModel.getConfigurationAuthority() == configurationArea.getConfigurationAuthority()) {
				// Der Bereich wird mit der größt möglichen Version für andere
				// zur Übernahme freigegeben
				configAreasAndVersions
						.add(new ConfigAreaAndVersion(configurationArea, configurationArea.getActivatableVersion()));
			}
		}

		KonfigAssistentErgebnis ergebnis = new KonfigAssistentErgebnis();
		ergebnis.setBetrachteteBereiche(configAreasAndVersions);
		ergebnis.setAreaDependencyCheckResult(new AreaDependencyChecker().checkAreas(configAreasAndVersions));

		final ConsistencyCheckResultInterface consistencyCheckResult = ((ConfigurationControl) dataModel)
				.releaseConfigurationAreasForActivationWithoutCAActivation(configAreasAndVersions);
		if (consistencyCheckResult.interferenceErrors()) {
			final String meldung = "Bei der Aktivierung durch andere ohne vorherige Aktivierung durch den KV wurden folgende Inkonsistenzen erkannt, die nicht zu einem Abbruch der Aktion geführt haben: "
					+ consistencyCheckResult;
			throw new Exception(meldung);
		}

		ergebnis.setConsistencyCheckResult(consistencyCheckResult);
		return ergebnis;

	}
}
