/*
 * Rahmenwerk-Plug-in "Konfigurationsassistent"
 *
 * Copyright (C) 2018 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weissenfelser Strasse 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 *
 * Copyright (c) 2011 SWARCO TRAFFIC SYSTEMS GmbH.
 * 
 * Contributors:
 * SWARCO TRAFFIC SYSTEMS GmbH - initial API and implementation
 *
 * Contact Information:
 * SWARCO TRAFFIC SYSTEMS GmbH
 * Adolf-Dambach-Straße
 * D-76571 Gaggenau
 * Phone: +49-7225-64-03
 * Fax: +49-7225-64-300
 * mailto: office.sts@swarco.de
 */

package de.bsvrz.buv.plugin.konfigass;

import java.time.LocalDateTime;
import java.util.concurrent.atomic.AtomicReference;
import java.util.logging.Level;

import org.eclipse.core.databinding.observable.Realm;
import org.eclipse.core.databinding.observable.map.ObservableMap;
import org.eclipse.core.databinding.observable.map.WritableMap;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.MultiStatus;
import org.eclipse.jface.databinding.swt.DisplayRealm;
import org.eclipse.swt.widgets.Display;

import de.bsvrz.buv.plugin.konfigass.handler.KonfigAssSichtHandler;
import de.bsvrz.buv.plugin.konfigass.log.MeldungsElement;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Speichert bisher empfangene Betriebsmeldungen. <br>
 * Hinweis: Diese Klasse ist als Singleton ausgelegt, damit auch zu einem
 * späteren Zeitpunkt, nach dem schließen des Sicht
 * {@link KonfigAssSichtHandler}, alle bereits erzeugten Meldungen wieder
 * angezeigt werden können.
 *
 * @author BitCtrl Systems GmbH, Uwe Peuker, HofmannJ (Dambach)
 *
 */
public final class MeldungenSpeicher {

	/**
	 * Speichert die einzigste Instanz der Klasse {@link MeldungenSpeicher}.
	 */
	private static MeldungenSpeicher instanz;

	private int maximaleAnzahlMultiMeldungen = 1000;

	/**
	 * Speichert die empfangenen Betriebsmeldungen und hält diese vorrätig.
	 */
	private final ObservableMap<LocalDateTime, MeldungsElement> meldungen;

	private static final Debug LOGGER = Debug.getLogger();

	private String konfiguration = "";

	/**
	 * Liefert die einzigste Instanz der Klasse {@link MeldungenSpeicher}.
	 *
	 * @return Einzigste Instanz der Klasse {@link MeldungenSpeicher}.
	 */
	public static synchronized MeldungenSpeicher getInstanz() {
		if (MeldungenSpeicher.instanz == null) {
			MeldungenSpeicher.instanz = new MeldungenSpeicher(DisplayRealm.getRealm(Display.getDefault()));
		}
		return MeldungenSpeicher.instanz;
	}

	/**
	 * Privater, nicht öffentlicher Konstruktor dieser Singletonklasse.
	 *
	 * @param maxAnzahlMultiMeldungen Setzt die maximale Anzahl von Meldungen, die
	 *                                gespeichert werden.
	 */
	private MeldungenSpeicher(Realm realm) {
		LOGGER.setLoggerLevel(Level.FINEST);
		meldungen = new WritableMap<>(realm);
	}

	public void hinzufuegenMultimeldung(final IStatus meldung) {

		LocalDateTime timeStamp = LocalDateTime.now();
		final String currentKonfiguration = konfiguration;

		Display.getDefault().asyncExec(() -> {
			MeldungsElement element = new MeldungsElement(meldung, timeStamp, currentKonfiguration);
			meldungen.put(timeStamp, element);
			
			if (meldungen.size() > maximaleAnzahlMultiMeldungen) {
				removeAeltesteMultimeldung();
			}
		});
	}

	@SuppressWarnings("unused")
	@Deprecated
	public void hinzufuegenMultimeldung(MultiStatus status, long zeit, boolean erwarteAusgabe) {
		hinzufuegenMultimeldung(status);
	}

	@SuppressWarnings("unused")
	@Deprecated
	public void setStatus(MultiStatus status) {
		// Tut nichts
	}

	public ObservableMap<LocalDateTime, MeldungsElement> getMeldungen() {
		return meldungen;
	}

	/**
	 * Entfernt die älteste Multimeldung in der Liste der Betriebstmeldungen.
	 *
	 */
	private void removeAeltesteMultimeldung() {
		
		final AtomicReference<LocalDateTime> toRemove = new AtomicReference<>();
		LocalDateTime timeStamp = LocalDateTime.MAX;

		for (LocalDateTime key : meldungen.keySet()) {
			if( timeStamp.isAfter(key)) {
				toRemove.set(key);
				timeStamp = key;
			}
		}

		if (toRemove.get() != null) {
			Display.getDefault().asyncExec(() -> meldungen.remove(toRemove.get()));
		}
	}

	public String getKonfiguration() {
		return konfiguration;
	}

	public void setKonfiguration(final String konfiguration) {
		this.konfiguration = konfiguration;
	}

	public void clear() {
		Display.getDefault().asyncExec(()->meldungen.clear());
	}
}
