/*
 * Rahmenwerk-Plug-in "Konfigurationsassistent"
 *
 * Copyright (C) 2018 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weissenfelser Strasse 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 *
 * Copyright (c) 2011 SWARCO TRAFFIC SYSTEMS GmbH.
 * 
 * Contributors:
 * SWARCO TRAFFIC SYSTEMS GmbH - initial API and implementation
 *
 * Contact Information:
 * SWARCO TRAFFIC SYSTEMS GmbH
 * Adolf-Dambach-Straße
 * D-76571 Gaggenau
 * Phone: +49-7225-64-03
 * Fax: +49-7225-64-300
 * mailto: office.sts@swarco.de
 */

package de.bsvrz.buv.plugin.konfigass;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintWriter;
import java.net.URL;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.MultiStatus;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;
import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import de.bsvrz.puk.config.main.managementfile.ManagementFile;

/**
 * Klasse zur Verwaltung von XML-Filtern, welche Listen von regulären Ausdrücken
 * enthalten. Liefert Funktionen zum Einlesen und Schreiben der entsprechenden
 * XML-Strukturen.
 *
 * @author BitCtrl Systems GmbH, Uwe Peuker, HofmannJ (Dambach)
 *
 */
class XmlElementKonfigVerwaltung {

	/** Bezeichnung eines Filterelementes. */
	static final String XML_ELEMENT_KB = "konfigurationsbereich";

	/** Regulärer Ausdruck innerhalb des Filterelementes. */
	static final String XML_ELEMENT_PID = "pid";

	/** Bezeichnung des Wurzelelementes der XML-Struktur. */
	private static final String XML_ELEMENT_WURZEL = "verwaltungsdaten";

	/** Bezeichnung eines Filterelementes. */
	private static final String XML_ELEMENT_KV = "konfigurationsverantwortlicher";

	/**
	 * JDOM Document, in dem die Verarbeitung erfolgt
	 */
	private Document document;

	/**
	 * JDOM Document, in dem die Verarbeitung erfolgt
	 */
	private static final String VERWALTUNG_SYSTEM_ID = "verwaltungsdaten.dtd";

	/**
	 * Wurzelelement
	 */
	private Element elementWurzel;

	public Element getElementWurzel() {
		return elementWurzel;
	}

	public void setElementWurzel(final Element elementWurzel) {
		this.elementWurzel = elementWurzel;
	}

	XmlElementKonfigVerwaltung() {
		elementWurzel = new Element(XmlElementKonfigVerwaltung.XML_ELEMENT_WURZEL);
		document = new Document(elementWurzel);
	}

	/**
	 * Liefert die zu dieser Struktur zugehörigen {@link ElementKonfig} nach ID
	 * sortiert zurück.
	 *
	 * @return Zugehörige XML-Filter
	 */
	public List<ElementKonfig> getElementKonfigBereich() {
		List<ElementKonfig> re = null;

		final List<Element> elements = elementWurzel.getChildren(XmlElementKonfigVerwaltung.XML_ELEMENT_KB);
		if ((elements != null) && (elements.size() > 0)) {
			re = new LinkedList<>();

			for (final Element element : elements) {
				final ElementKonfig elementKonfig = new ElementKonfig(element);
				re.add(elementKonfig);
			}
		}

		/*
		 * Sortieren nach ID
		 */
		if ((re != null) && (re.size() > 1)) {
			Collections.sort(re);
		}

		return re;
	}

	/**
	 * Liefert die zu dieser Struktur zugehörigen {@link ElementKonfig} nach ID
	 * sortiert zurück.
	 *
	 * @return Zugehörige XML-Filter
	 */
	public ElementKonfig getElementKonfigVerantwortlicher() {
		ElementKonfig re = null;

		final List<Element> elements = elementWurzel.getChildren(XmlElementKonfigVerwaltung.XML_ELEMENT_KV);

		if ((elements != null) && (elements.size() == 1)) {
			re = new ElementKonfig(elements.get(0));
		}

		return re;
	}

	void entferneElementKonfig(final ElementKonfig elementKonfig) {
		/*
		 * Speichert zu löschende Elemente
		 */
		final List<Element> loeschListe = new LinkedList<>();

		/*
		 * Suchen der Löschelemente
		 */
		final Iterator<Element> itPruefen = elementWurzel.getChildren(XmlElementKonfigVerwaltung.XML_ELEMENT_KB)
				.iterator();
		while (itPruefen.hasNext()) {
			final Element element = itPruefen.next();
			final ElementKonfig ek = new ElementKonfig(element);
			if (ek.getBezeichnung().equals(elementKonfig.getBezeichnung())) {
				loeschListe.add(element);
			}
		}

		/*
		 * Löschen der Elemente aus der Löschliste
		 */
		final Iterator<Element> itLoeschen = loeschListe.iterator();
		while (itLoeschen.hasNext()) {
			final Element loeschElement = itLoeschen.next();
			elementWurzel.removeContent(loeschElement);
		}

	}

	/**
	 * Liest eine XML-Datei ein
	 *
	 * @param dateiPfad Kompletter Pfad zur Datei
	 * @throws IllegalArgumentException Wird geworfen, wenn die Datei nicht
	 *                                  existiert
	 */
	void einlesenXmlAusDatei(final String dateiPfad) {
		final SAXBuilder builder = new SAXBuilder();

		final File file = new File(dateiPfad);
		if (file.exists()) {
			try {
				builder.setValidation(false);

				builder.setEntityResolver(new EntityResolver() {

					@Override
					public InputSource resolveEntity(final String publicId, final String systemId)
							throws SAXException, IOException {
						if ((systemId != null) && systemId.endsWith(XmlElementKonfigVerwaltung.VERWALTUNG_SYSTEM_ID)) {
							final URL url = ManagementFile.class
									.getResource(XmlElementKonfigVerwaltung.VERWALTUNG_SYSTEM_ID);

							if (url != null) {

								final String ef = url.toExternalForm();

								if (ef.length() > 0) {
									return new InputSource(ef);
								}
							}
						}
						final MultiStatus status = new MultiStatus(KonfigurationsAssistent.PLUGIN_ID, IStatus.ERROR,
								"Fehler beim Auflösen der DTD-Datei:" + XmlElementKonfigVerwaltung.VERWALTUNG_SYSTEM_ID,
								null);
						MeldungenSpeicher.getInstanz().hinzufuegenMultimeldung(status);

						return null;
					}
				});

				document = builder.build(file);
				elementWurzel = document.getRootElement();
			} catch (final JDOMException e) {
				e.printStackTrace();
			} catch (final IOException e) {
				e.printStackTrace();
			}

		} else {
			throw new IllegalArgumentException("Die Datei " + dateiPfad + " existiert nicht!");
		}
	}

	/**
	 * Liefert eine formatierte Ausgabe der erzeugten XML-Struktur. Setzt das
	 * Dirty-Flag auf <code>false</code>.
	 *
	 * @return Formatierte XML-Struktur
	 */
	String ausgabeXmlFormatiert() {
		String returnString = null;
		if (document == null) {
			document = new Document(elementWurzel);
		}

		final Format prettyFormat = Format.getPrettyFormat();
		final XMLOutputter out = new XMLOutputter(prettyFormat);
		final Format f = out.getFormat();
		// f.setEncoding(XmlBezeichnerMenuSymbol.xmlCodierung);
		f.setEncoding("ISO-8859-1");
		out.setFormat(f);

		returnString = out.outputString(document);

		return returnString;
	}

	/**
	 * Schreibt einen String in eine Datei
	 *
	 * @param dateiPfad      Kompletter Pfad zur Datei, in die geschrieben werden
	 *                       soll
	 * @param schreibString  String, der in die Datei geschrieben werden soll
	 * @param ueberschreiben true == Überschreiben der Datei, falls schon vorhanden
	 * @throws IllegalArgumentException Wir geworfen, wenn die Datei existiert und
	 *                                  nicht überschrieben werden kann
	 */
	void schreibeXMLStringInDatei(final String dateiPfad, final String schreibString, final boolean ueberschreiben) {
		final File file = new File(dateiPfad);
		// File existiert, auf Überschreiben prüfen
		if (file.exists() && ueberschreiben) {
			try (PrintWriter pw = new PrintWriter(file)) {
				pw.print(schreibString);
			} catch (final FileNotFoundException e) {
				e.printStackTrace();
			}
		}
		// File existiert nicht und kann einfach angelegt werden
		else if (!file.exists()) {
			try (PrintWriter pw = new PrintWriter(file)) {
				pw.print(schreibString);
			} catch (final FileNotFoundException e) {
				e.printStackTrace();
			}
		}
		// File existiert und ueberschreiben == flase
		else {
			throw new IllegalArgumentException("Die Datei " + dateiPfad + " existiert und ueberschreiben == flase");
		}
	}

	@Override
	public String toString() {
		final StringBuilder re = new StringBuilder();
		re.append("===XmlFilterVerwaltung===\n");

		final Iterator<Element> it = elementWurzel.getChildren(XmlElementKonfigVerwaltung.XML_ELEMENT_KB).iterator();
		while (it.hasNext()) {
			final Element element = it.next();
			final ElementKonfig fi = new ElementKonfig(element);
			re.append("FI: ");
			re.append(fi);
		}

		re.append("=========================");
		return re.toString();
	}
}
