/*
 * Rahmenwerk-Plug-in "Konfigurationsassistent"
 *
 * Copyright (C) 2018 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weissenfelser Strasse 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 *
 * Copyright (c) 2011 SWARCO TRAFFIC SYSTEMS GmbH.
 * 
 * Contributors:
 * SWARCO TRAFFIC SYSTEMS GmbH - initial API and implementation
 *
 * Contact Information:
 * SWARCO TRAFFIC SYSTEMS GmbH
 * Adolf-Dambach-Straße
 * D-76571 Gaggenau
 * Phone: +49-7225-64-03
 * Fax: +49-7225-64-300
 * mailto: office.sts@swarco.de
 */

package de.bsvrz.buv.plugin.konfigass.datenModellUpdateListener;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import de.bsvrz.dav.daf.main.config.SystemObject;

/**
 * Die Klasse kapselt die Verwaltungsmethoden für die Listener-Funktionalität
 * des {@link DatenModellUpdateListener}. Diese Klasse implementiert das
 * Singleton-Pattern und stellt dadurch sicher dass der Listener nur einmal
 * Instanziiert wird.
 *
 * @author BitCtrl Systems GmbH, Uwe Peuker, Pittner (Dambach)
 *
 */
public class DatenModellUpdateListener {

	/**
	 * Liste der {@link IDatenModellUpdateListener}, die über Aktualisierungen
	 * informiert werden möchten.
	 */
	private final List<IDatenModellUpdateListener> datenModellUpdateListener = new LinkedList<>();

	/**
	 * Singleton-Instanz
	 */
	private static DatenModellUpdateListener instanz;

	/**
	 * Liefert die Singletoninstanz.
	 *
	 * @return Instanz dieser Klasse.
	 */
	public static synchronized DatenModellUpdateListener getInstanz() {
		if (DatenModellUpdateListener.instanz == null) {
			DatenModellUpdateListener.instanz = new DatenModellUpdateListener();
		}
		return DatenModellUpdateListener.instanz;
	}

	/**
	 * Fügt einen Listener zur Liste der {@link DatenModellUpdateListener} hinzu.
	 *
	 * @param listener
	 */
	public void addDatenModellUpdateListener(final IDatenModellUpdateListener listener) {
		if (!datenModellUpdateListener.contains(listener)) {
			datenModellUpdateListener.add(listener);
		}
	}

	/**
	 * Benachrichtigt alle registrierten {@link IDatenModellUpdateListener} darüber,
	 * dass neue Daten vorhanden sind.
	 *
	 * @param src
	 *            die Quelle, die das Ereignis auslöst
	 * @param listeAktualisiert
	 *            Liste der aktualisierten SystemObjekte
	 * @param listeGeloescht
	 *            Liste der geloeschten SystemObjekte
	 * @param listeNeu
	 *            Liste der neu erstellten SystemObjekte
	 */
	public void notifyDatenModellUpdateListener(final Object src, final List<SystemObject> listeAktualisiert,
			final List<SystemObject> listeGeloescht, final List<SystemObject> listeNeu) {
		final DatenModellUpdateEvent event = new DatenModellUpdateEvent(src, listeAktualisiert, listeGeloescht,
				listeNeu);

		final Iterator<IDatenModellUpdateListener> it = datenModellUpdateListener.iterator();
		while (it.hasNext()) {
			final IDatenModellUpdateListener listener = it.next();
			listener.datenModellUpdate(event);
		}
	}

	/**
	 * Entfernt einen {@link IDatenModellUpdateListener} von der Liste der Zuhörer.
	 *
	 * @param listener
	 *            Zu entfernender {@link IDatenModellUpdateListener}.
	 */
	public void removeDatenModellUpdateListener(final IDatenModellUpdateListener listener) {
		if (datenModellUpdateListener.contains(listener)) {
			datenModellUpdateListener.remove(listener);
		}
	}

}
