/*
 * Rahmenwerk-Plug-in "Konfigurationsassistent"
 *
 * Copyright (C) 2018 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weissenfelser Strasse 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 *
 * Copyright (c) 2011 SWARCO TRAFFIC SYSTEMS GmbH.
 * 
 * Contributors:
 * SWARCO TRAFFIC SYSTEMS GmbH - initial API and implementation
 *
 * Contact Information:
 * SWARCO TRAFFIC SYSTEMS GmbH
 * Adolf-Dambach-Straße
 * D-76571 Gaggenau
 * Phone: +49-7225-64-03
 * Fax: +49-7225-64-300
 * mailto: office.sts@swarco.de
 */

package de.bsvrz.buv.plugin.konfigass.log;


import java.io.FileWriter;
import java.io.IOException;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map.Entry;
import java.util.SortedMap;
import java.util.TreeMap;

import org.eclipse.core.databinding.observable.map.IMapChangeListener;
import org.eclipse.core.databinding.observable.map.MapChangeEvent;
import org.eclipse.core.databinding.observable.map.MapDiff;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.ITableLabelProvider;
import org.eclipse.jface.viewers.ITreeContentProvider;
import org.eclipse.jface.viewers.LabelProvider;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.widgets.FileDialog;

import de.bsvrz.buv.plugin.konfigass.KonfigurationsAssistent;
import de.bsvrz.buv.plugin.konfigass.MeldungenSpeicher;

/**
 * Klasse verwaltet das GUI 'Protokollmeldungen'.
 *
 * @author BitCtrl Systems GmbH, Uwe Peuker, Liliya Givorgizova (LG, inovat)
 *
 */
public class ProtokollGUIVerwaltung {

	/** Liste der Meldungselementen. */
	private final SortedMap<LocalDateTime, MeldungsElement> listeMeldungElementen = new TreeMap<>();

	/** GUI-Instanz. */
	private final ProtokollGUI gui;

	private IMapChangeListener<LocalDateTime, MeldungsElement> changeListener;

	/**
	 * Erzeugt eine Instanz der Klasse ProtokollGUIVerwaltung.
	 *
	 * @param gui GUI-Instanz.
	 */
	protected ProtokollGUIVerwaltung(final ProtokollGUI gui) {
		this.gui = gui;

		listeMeldungElementen.putAll(MeldungenSpeicher.getInstanz().getMeldungen());
		initBaum();

		changeListener = new IMapChangeListener<LocalDateTime, MeldungsElement>() {

			@Override
			public void handleMapChange(MapChangeEvent<? extends LocalDateTime, ? extends MeldungsElement> event) {
				MapDiff<? extends LocalDateTime, ? extends MeldungsElement> diff = event.diff;
				if (diff.getRemovedKeys().isEmpty() && diff.getChangedKeys().isEmpty()) {
					for (LocalDateTime zeit : diff.getAddedKeys()) {
						listeMeldungElementen.put(zeit, diff.getNewValue(zeit));
					}
				} else {
					listeMeldungElementen.clear();
					listeMeldungElementen.putAll(MeldungenSpeicher.getInstanz().getMeldungen());
				}
				initBaum();
			}
		};
		MeldungenSpeicher.getInstanz().getMeldungen().addMapChangeListener(changeListener);

	}

	// ~ METHODEN ==============================================================

	protected void btnLoeschenSelektiert() {
		if (listeMeldungElementen.size() > 0) {
			if (MessageDialog.openQuestion(gui.getShell(), "Frage",
					"Möchten Sie alle Meldungen aus dem Protokoll löschen?\n\n"
							+ "ACHTUNG: der Prozess kann nicht rückgängig gemacht werden!")) {
				loescheAlleMeldungen();
			}
		}
	}

	/** Button 'Speichern' wird selektiert. */
	protected void btnSpeichernSelektiert() {
		waehleDateiUndSpeichereAlleMeldungen();
	}

	/**
	 * Ermittle aller Meldungen im Textform.
	 *
	 * @return Aller Meldungen im Textform.
	 */
	private String ermittleTextAllerMeldungen() {
		final StringBuilder alleMeldungen = new StringBuilder();

		for (final Entry<LocalDateTime, MeldungsElement> entry : listeMeldungElementen.entrySet()) {
			alleMeldungen.append(erstelleMeldungsTextRekursiv(entry.getValue(), 0));
		}

		return alleMeldungen.toString();
	}

	/**
	 * Erstelle den Text für die Meldung. Besteht aus der Meldung und Exception.
	 *
	 * @param me MeldungsElement.
	 *
	 * @return Text für die Meldung.
	 */
	private String erstelleMeldungsText(final MeldungsElement me) {
		final StringBuffer meldung = new StringBuffer();

		final String message = me.getMeldung().getMessage();
		meldung.append(System.getProperty("line.separator"));
		meldung.append(message);

		// hole die ExceptionInformation:
		// noinspection ThrowableResultOfMethodCallIgnored
		final Throwable e = me.getMeldung().getException();

		if (e != null) {
			meldung.append("\n\n");
			meldung.append(e.toString()).append("\n\n");

			// noinspection ThrowableResultOfMethodCallIgnored
			for (final StackTraceElement el : me.getMeldung().getException().getStackTrace()) {
				meldung.append(el.toString()).append("\n");
			}
		}

		meldung.append("\n");
		return meldung.toString();
	}

	/**
	 * Erstelle den Text für die Meldung und ihre Kinder. Besteht aus der Meldungen
	 * und Exception.
	 *
	 * @param me MeldungsElement.
	 * @param i  Definiert die Ebene.
	 *
	 * @return Text für die Meldung und ihre Kinder.
	 */
	private String erstelleMeldungsTextRekursiv(final MeldungsElement me, final int i) {

		// Initialisiere die Abstand- und Trennzeichen:
		final String einAbstand = "    ";
		final int laenge = 120;
		final String trennZeichen;

		if (me.getListeKinder().size() > 0) {
			trennZeichen = "=";
		} else {
			trennZeichen = "-";
		}

		// Erstelle den Abstand und die Trennlinie:
		final StringBuffer abstand = new StringBuffer();

		for (int j = 0; j < i; j++) {
			abstand.append(einAbstand);
		}

		final StringBuffer trennLinie = new StringBuffer();

		for (int j = 0; j < (laenge - (i * einAbstand.length())); j++) {
			trennLinie.append(trennZeichen);
		}

		// Füge die Trennlinie zu der Ausgabe zu:
		final StringBuffer meldung = new StringBuffer();

		meldung.append(String.format("%s%s%n", abstand, trennLinie));

		// Füge den Status zu:
		if (me.getKonfiguration() != null) {
			meldung.append(String.format("%s%s%s%s%n", abstand, Log.ermittleStatusText(me.getMeldung()), "         ",
					me.getKonfiguration()));
		} else {
			meldung.append(String.format("%s%s%n", abstand, Log.ermittleStatusText(me.getMeldung())));
		}

		// Füge die Trennlinie zu:
		meldung.append(String.format("%s%s%n", abstand, trennLinie));

		// Füge den einzeln MeldungsText zu:
		final String[] arZeile = erstelleMeldungsText(me).split("\n");

		for (final String zeile : arZeile) {
			meldung.append(String.format("%s%s%n", abstand, zeile));
		}

		// Füge den MeldungsText der Kinder zu:
		for (final MeldungsElement kind : me.getListeKinder()) {
			meldung.append(erstelleMeldungsTextRekursiv(kind, i + 1));
		}

		return meldung.toString();
	}

	/** Initialisiere den MeldungsBaum. */
	private void initBaum() {
		gui.getDisplay().asyncExec(() -> gui.getTreeViewerMeldungen().setInput(listeMeldungElementen));
	}

	/** Lösche den Inhalt des Protokolls. */
	private void loescheAlleMeldungen() {
		MeldungenSpeicher.getInstanz().clear();
		listeMeldungElementen.clear();
		initBaum();
	}

	/**
	 * Speichere alle Meldungen in die ASCII-Datei.
	 *
	 * @param dateiName Name der ASCII-Datei.
	 *
	 * @return true, falls die Methode erfolgreich beendet wurde, false - sonst.
	 */
	private boolean speichereAlleMeldungenInDieDatei(final String dateiName) {
		boolean erfolgreich = true;

		try (FileWriter file = new FileWriter(dateiName)) {
			file.write(ermittleTextAllerMeldungen());
		} catch (final IOException e) {
			erfolgreich = false;
			new MultiMeldung(KonfigurationsAssistent.PLUGIN_ID, IStatus.ERROR,
					"Die Datei <" + dateiName + "> kann nicht erstellt/überschrieben werden!", e).zeige();
		}

		return erfolgreich;
	}

	/**
	 * Öffne den Auswahldialog für die Datei und speichere alle Meldungen in die
	 * ausgewählte Datei.
	 */
	private void waehleDateiUndSpeichereAlleMeldungen() {

		// Auswahldialog für die Datei anzeigen:
		final FileDialog dialog = new FileDialog(gui.getShell(), SWT.SAVE);

		dialog.setFileName("Protokoll.txt");
		dialog.setFilterExtensions(new String[] { "*.txt" });
		dialog.setText("Protokolldatei zum Speichern wählen");

		final String datei = dialog.open();

		if (datei != null) {
			// Meldungen in die Datei speichern:
			if (!speichereAlleMeldungenInDieDatei(datei)) {
				MessageDialog.openError(gui.getShell(), "Fehler",
						"Es ist ein Fehler beim Speichern der Meldungen in die Datei aufgetreten (s. Meldung unten) !");
			}
		}
	}

	/** ContentProvider für den MeldungsBaum. */
	static class LogTreeContentProvider implements ITreeContentProvider {
		private static final Object[] EMPTY_OBJECT_ARRAY = new Object[0];

		@Override
		public void dispose() {
			// keine Ressourcen freizugeben
		}

		@Override
		public Object[] getChildren(final Object parentElement) {
			if (parentElement instanceof MeldungsElement) {
				final MeldungsElement me = (MeldungsElement) parentElement;

				return me.getListeKinder().toArray(new MeldungsElement[me.getListeKinder().size()]);
			}

			return LogTreeContentProvider.EMPTY_OBJECT_ARRAY;
		}

		@Override
		public Object[] getElements(final Object inputElement) {

			final List<MeldungsElement> list = new ArrayList<>();
			if( inputElement instanceof SortedMap) {
				for( Object value : ((SortedMap<?,?>) inputElement).values()) {
					if( value instanceof MeldungsElement) {
						list.add((MeldungsElement) value);
					}
				}
			}
			
			return list.toArray(new MeldungsElement[list.size()]);
		}

		@Override
		public Object getParent(final Object element) {
			if (element instanceof MeldungsElement) {
				final MeldungsElement me = (MeldungsElement) element;

				return me.getVorfahr();
			}

			return null;
		}

		@Override
		public boolean hasChildren(final Object element) {
			if (element instanceof MeldungsElement) {
				final MeldungsElement me = (MeldungsElement) element;
				if (me.getMeldung() != null) {
					return me.getMeldung().getChildren().length > 0;
				}
			}
			return false;
		}

		@Override
		public void inputChanged(final Viewer viewer, final Object oldInput, final Object newInput) {
			// der übergebene Input wird nicht verwendet
		}
	}

	/** LabelProvider für den MeldungsBaum. */
	static class LogTreeLabelProvider extends LabelProvider implements ITableLabelProvider {
		private static final Image INFO_IMAGE = KonfigurationsAssistent.getDefault()
				.getImage("icons/pde/info_st_obj.gif");
		private static final Image OK_IMAGE = KonfigurationsAssistent.getDefault().getImage("icons/pde/ok_st_obj.gif");
		private static final Image ERROR_IMAGE = KonfigurationsAssistent.getDefault()
				.getImage("icons/pde/error_st_obj.gif");
		private static final Image WARNING_IMAGE = KonfigurationsAssistent.getDefault()
				.getImage("icons/pde/warning_st_obj.gif");

		// ~ GET METHODEN ======================================================

		@Override
		public Image getColumnImage(final Object element, final int columnIndex) {
			if ((columnIndex == 0) && (element instanceof MeldungsElement)) {
				final MeldungsElement me = (MeldungsElement) element;

				if (me.getMeldung() != null) {
					switch (me.getMeldung().getSeverity()) {
					case IStatus.ERROR:
						return LogTreeLabelProvider.ERROR_IMAGE;

					case IStatus.INFO:
						return LogTreeLabelProvider.INFO_IMAGE;

					case IStatus.OK:
						return LogTreeLabelProvider.OK_IMAGE;

					case IStatus.WARNING:
						return LogTreeLabelProvider.WARNING_IMAGE;

					default:
						break;
					}
				}
			}

			return null;
		}

		@Override
		public String getColumnText(final Object element, final int columnIndex) {
			String returnText = "";

			if (element instanceof MeldungsElement) {
				final MeldungsElement me = (MeldungsElement) element;

				if (me.getMeldung() != null) {

					switch (columnIndex) {
					case 0:

						// Falls die Meldung mehr Zeilen hat,
						// wird nur die 1.Zeile im Baum angezeigt.
						String text = "";
						if (me.getMeldung() != null) {
							text = me.getMeldung().getMessage();
							final String[] arTest = text.split("\n");

							if (arTest.length > 1) {
								text = arTest[0] + "...";
							}

						}
						return text;

					case 1:
						returnText = me.getMeldung().getPlugin();
						break;

					case 2:
						if (me.getMeldungsZeit() != LocalDateTime.MIN) {
							returnText = LogTreeLabelProvider.getDatumZeitAlsString(me.getMeldungsZeit());
						}
						break;
					case 3:
						if (me.getKonfiguration() != null) {
							returnText = me.getKonfiguration();
						}
						break;

					default:
						break;
					}
				}

			}

			return returnText;
		}

		/**
		 * Stellt Datum und Uhrzeit als String dar.
		 *
		 * @param datumLong - Anzahl der Millisekunden.
		 *
		 * @return Datum als String "dd.mm.yyyy hh:mm:ss".
		 */
		private static String getDatumZeitAlsString(final LocalDateTime zeit) {
			return DateTimeFormatter.ofPattern("dd.MM.yyyy HH:mm:ss").format(zeit);
		}
	}

	public void dispose() {
		MeldungenSpeicher.getInstanz().getMeldungen().removeMapChangeListener(changeListener);
	}
}
