/*
 * Rahmenwerk-Plug-in "Konfigurationsassistent"
 *
 * Copyright (C) 2018 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weissenfelser Strasse 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 *
 * Copyright (c) 2011 SWARCO TRAFFIC SYSTEMS GmbH.
 * 
 * Contributors:
 * SWARCO TRAFFIC SYSTEMS GmbH - initial API and implementation
 *
 * Contact Information:
 * SWARCO TRAFFIC SYSTEMS GmbH
 * Adolf-Dambach-Straße
 * D-76571 Gaggenau
 * Phone: +49-7225-64-03
 * Fax: +49-7225-64-300
 * mailto: office.sts@swarco.de
 */

package de.bsvrz.buv.plugin.konfigass.managementfile;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Document;
import org.w3c.dom.DocumentType;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import de.bsvrz.puk.config.main.managementfile.ConfigurationAreaManagementInfo;
import de.bsvrz.puk.config.main.managementfile.VersionInfo;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.filelock.FileLock;
import de.bsvrz.sys.funclib.xmlSupport.CountingErrorHandler;

public class MyManagementFile implements MyConfigurationManagementFile {

	/** DebugLogger für Debug-Ausgaben */
	private static final Debug LOGGER = Debug.getLogger();

	/** Repräsentiert die vollständige XML-Datei. */
	private final Document xmlDocument;

	/** Datei, wo die Verwaltungsdaten abgespeichert werden sollen. */
	private final File xmlFile;

	/** Speichert die Einträge zu den Verwaltungsdaten in dieser Liste. */
	private final List<MyConfigurationAreaManagementInfo> areaManagementInfos;

	/** Das Format des Zeitstempels für die Versionsaktivierung. */
	private DateFormat dateFormat;

	/** Speichert die Basis der Verzeichnisse für die Konfigurationsbereiche. */
	private final URI uriBase;

	private final FileLock managementFileLock;

	private File objectSetBaseDirectory;

	/**
	 * Der Konstruktor lädt die angegebene XML-Datei mit den Verwaltungsdaten der
	 * Konfiguration, validiert gegen die dazugehörende DTD bzw. erzeugt eine neue
	 * Datei.
	 *
	 * @param source die Verwaltungsdatei
	 */
	public MyManagementFile(final File source) {
		if (source == null) {
			throw new IllegalArgumentException("Es wurde kein File-Objekt angegeben.");
		}

		try {
			xmlFile = source.getCanonicalFile();
		} catch (final IOException e) {
			throw new IllegalArgumentException(e);
		}

		// Prüfen, ob sich die Datei bereits im Zugriff befindet, wenn nicht
		// wird diese für andere gesperrt
		managementFileLock = new FileLock(xmlFile);
		try {
			managementFileLock.lock();
		} catch (final IOException e) {
			final String errorMessage = "Die lock-Datei für die Datei " + xmlFile.getAbsolutePath()
					+ " konnte nicht erzeugt werden";
			e.printStackTrace();
			MyManagementFile.LOGGER.error(errorMessage, e);
			throw new IllegalStateException(errorMessage, e);
		}

		try {
			// Verwaltungsdaten einladen!
			final CountingErrorHandler errorHandler = new CountingErrorHandler();
			final DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
			factory.setNamespaceAware(true);

			// die Validierung der XML-Datei anhand der DTD durchführen
			factory.setValidating(true);
			factory.setAttribute("http://xml.org/sax/features/validation", Boolean.TRUE);

			try {
				final DocumentBuilder builder = factory.newDocumentBuilder();

				if (xmlFile.exists()) {
					if (!xmlFile.isFile()) {
						throw new IllegalArgumentException(
								"Das übergebene File-Objekt ist keine Datei: " + xmlFile.toString());
					}
					MyManagementFile.LOGGER.config("Datei wird eingelesen", xmlFile);
					try {
						builder.setErrorHandler(errorHandler);

						final ManagementEntityResolver resolver = new ManagementEntityResolver();

						builder.setEntityResolver(resolver);
						xmlDocument = builder.parse(xmlFile); // evtl. mittels
						// BufferedInputStream
						// cachen
						errorHandler.printSummary();
						if (errorHandler.getErrorCount() > 0) {
							throw new IllegalStateException(errorHandler.getErrorCount()
									+ " Fehler beim Parsen der XML-Datei " + xmlFile.toString());
						}
					} catch (final Exception ex) {
						final String errorMessage = "Die Verwaltungdaten der Konfiguration konnten nicht eingelesen werden: "
								+ xmlFile.toString();
						throw new IllegalStateException(errorMessage, ex);
					}
				} else {
					xmlDocument = builder.getDOMImplementation().createDocument(null, "verwaltungsdaten", null);
					try (BufferedOutputStream outputStream = new BufferedOutputStream(new FileOutputStream(xmlFile))) {
						save(outputStream);
					}
				}
				objectSetBaseDirectory = xmlFile.getParentFile();
				uriBase = objectSetBaseDirectory.toURI();
				MyManagementFile.LOGGER.config("Verzeichnisbasis für die Konfigurationsbereiche", uriBase.toString());
			} catch (final ParserConfigurationException ex) {
				final String errorMessage = "Die Verwaltungdaten der Konfiguration konnten nicht eingelesen werden: "
						+ xmlFile.toString();
				MyManagementFile.LOGGER.error(errorMessage);
				throw new IllegalArgumentException(errorMessage, ex);
			} catch (final IOException ex) {
				final String errorMessage = "Es konnte keine neue Verwaltungsdatei erstellt werden";
				MyManagementFile.LOGGER.error(errorMessage, ex);
				throw new IllegalStateException(errorMessage, ex);
			}

			// Konfigurationsbereiche eingelesen
			areaManagementInfos = getAllManagementInfos();

			MyManagementFile.LOGGER.config("Verwaltungsdaten der Konfiguration wurden vollständig eingelesen.");
		} catch (final RuntimeException e) {
			// Falls im Konstruktor nach dem Lock der Verwaltungsdatei eine
			// Exception auftritt, dann wird ein Unlock
			// ausgeführt und die Exception weitergegeben.
			managementFileLock.unlock();
			throw e;
		}
	}

	/**
	 * Interne Methode, die die Verwaltungseinträge zu den Konfigurationsbereichen
	 * einlädt und gesammelt in einer Liste zurückgibt.
	 *
	 * @return eine Liste mit den Verwaltungseinträgen der Konfigurationsbereiche
	 */
	private List<MyConfigurationAreaManagementInfo> getAllManagementInfos() {
		final List<MyConfigurationAreaManagementInfo> resultList = new LinkedList<>();
		synchronized (xmlDocument) {
			final Element xmlRoot = xmlDocument.getDocumentElement();
			final NodeList entryList = xmlRoot.getElementsByTagName("konfigurationsbereich");
			for (int i = 0; i < entryList.getLength(); i++) {
				final Element element = (Element) entryList.item(i);
				final MyConfigAreaManagementInfo info = new MyConfigAreaManagementInfo(element.getAttribute("pid"),
						false);
				// Versionen einlesen
				final NodeList versionList = element.getElementsByTagName("version");
				for (int j = 0; j < versionList.getLength(); j++) {
					final Element version = (Element) versionList.item(j);
					final int nr = Integer.parseInt(version.getAttribute("nr"));
					String time = version.getAttribute("zeitpunkt");
					if (time.isEmpty()) {
						// diese Version wird aktiviert und erhält die aktuelle
						// Uhrzeit als Zeitstempel
						if (dateFormat == null) {
							dateFormat = new SimpleDateFormat("dd.MM.yyyy HH:mm:ss,SSS");
						}
						time = dateFormat.format(new Date(System.currentTimeMillis()));
						version.setAttribute("zeitpunkt", time);
						info.setNewVersionActivated(true);
					}
					final VersionInformation versionInformation = new VersionInformation((short) nr, time);
					// letzte Version ist die aktive Version
					if (j == (versionList.getLength() - 1)) {
						info.setActiveVersion(versionInformation);
					}
					info.addVersionInfo(versionInformation);
				}
				resultList.add(info);
			}
		}
		return resultList;
	}

	/**
	 * Gibt die Pid des Konfigurationsverantwortlichen der Konfiguration zurück.
	 *
	 * @return die Pid des Konfigurationsverantwortlichen der Konfiguration
	 */
	@Override
	public String getConfigurationAuthority() {
		synchronized (xmlDocument) {
			final Element xmlRoot = xmlDocument.getDocumentElement();
			final NodeList authorityList = xmlRoot.getElementsByTagName("konfigurationsverantwortlicher");
			if (authorityList.getLength() > 0) {
				final Element element = (Element) authorityList.item(0);
				return element.getAttribute("pid");
			}
			return "";
		}
	}

	/**
	 * Speichert die Pid des Konfigurationsverantwortlichen der Konfiguration ab.
	 *
	 * @param pid die Pid des Konfigurationsverantwortlichen
	 */
	@Override
	public void setConfigurationAuthority(final String pid) {
		synchronized (xmlDocument) {
			final Element xmlRoot = xmlDocument.getDocumentElement();
			final NodeList authorityList = xmlRoot.getElementsByTagName("konfigurationsverantwortlicher");
			if (authorityList.getLength() > 0) {
				// es gibt einen Eintrag
				final Element element = (Element) authorityList.item(0);
				element.setAttribute("pid", pid);
			} else {
				// ein neuer Eintrag muss hinzugefügt werden
				final Element element = xmlDocument.createElement("konfigurationsverantwortlicher");
				element.setAttribute("pid", pid);
				xmlDocument.getDocumentElement().appendChild(element);
			}
		}
	}

	@Override
	public List<MyConfigurationAreaManagementInfo> getAllMyConfigurationAreaManagementInfos() {
		return areaManagementInfos;
	}

	/**
	 * Gibt einen Eintrag aus den Verwaltungsdaten zu der angegebenen Pid eines
	 * Konfigurationsbereichs zurück. Falls es keinen Eintrag gibt wird
	 * <code>null</code> zurückgegeben.
	 *
	 * @param configurationAreaPid Pid eines Konfigurationsbereichs
	 *
	 * @return Eintrag aus den Verwaltungsdaten zu einem Konfigurationsbereich oder
	 *         <code>null</code>, falls kein Eintrag vorhanden ist
	 */
	@Override
	public ConfigurationAreaManagementInfo getConfigurationAreaManagementInfo(final String configurationAreaPid) {
		synchronized (areaManagementInfos) {
			for (final ConfigurationAreaManagementInfo configurationAreaManagementInfo : areaManagementInfos) {
				if (configurationAreaManagementInfo.getPid().equals(configurationAreaPid)) {
					return configurationAreaManagementInfo;
				}
			}
		}
		return null; // kein Eintrag vorhanden
	}

	/**
	 * Fügt einen neuen Eintrag eines Konfigurationsbereichs ans Ende der
	 * Verwaltungsdatei ein.
	 *
	 * @param pid Pid des Konfigurationsbereichs, welcher zu den Verwaltungsdaten
	 *            hinzugefügt werden soll
	 *
	 * @return der Eintrag in den Verwaltungsdaten zu einem Konfigurationsbereich
	 */
	@Override
	public ConfigurationAreaManagementInfo addConfigurationAreaManagementInfo(final String pid) {
		// Objekt erzeugen, in XML einfügen und in die Liste eintragen
		// erst prüfen, ob zu dieser Pid bereits ein Eintrag vorhanden ist
		synchronized (areaManagementInfos) {
			for (final MyConfigurationAreaManagementInfo configurationAreaManagementInfo : areaManagementInfos) {
				if (configurationAreaManagementInfo.getPid().equals(pid)) {
					throw new IllegalArgumentException(
							"Zu dieser Pid '" + pid + "' gibt es bereits einen Eintrag in den Verwaltungsdaten.");
				}
			}
		}
		final MyConfigurationAreaManagementInfo info = new MyConfigAreaManagementInfo(pid, true);
		synchronized (areaManagementInfos) {
			areaManagementInfos.add(info);
		}
		return info;
	}

	/** Speichert die Verwaltungsdaten ab. */
	@Override
	public void save() throws IOException {
		// Verwaltungsdaten in einer temporären Datei speichern
		final File parentFile = xmlFile.getParentFile();

		final File newFile = File.createTempFile("verwaltungsdaten", ".xml", parentFile);
		try (BufferedOutputStream outputStream = new BufferedOutputStream(new FileOutputStream(newFile))) {
			save(outputStream);
		} catch (final IOException ex) {
			MyManagementFile.LOGGER.error("Fehler beim Speichern der Verwaltungsdaten in der temporären Datei",
					newFile);
			throw new IOException("Fehler beim Speichern der Verwaltungsdaten " + ex.getMessage());
		}
		// Backup der bisherigen Datei erstellen
		final File backupFile = new File(parentFile, xmlFile.getName() + ".old");
		backupFile.delete(); // löscht eine evtl. bestehende alte Datei

		if (xmlFile.renameTo(backupFile)) {
			// temporäre Datei umbenennen zur Verwaltungsdatei
			if (!newFile.renameTo(xmlFile)) {
				final String errorMessage = "Temporäre Verwaltungsdatei konnte nicht in " + xmlFile.getName()
						+ " umbenannt werden.";
				MyManagementFile.LOGGER.error(errorMessage);
				throw new IOException(errorMessage);
			}
			if (newFile.exists()) {
				MyManagementFile.LOGGER
						.warning("Datei " + newFile + " existiert immer noch, obwohl sie eigentlich umbenannt wurde.");
			}
		} else {
			final String errorMessage = "Beim Sichern der Verwaltungsdaten konnte das Backup der bisherigen Verwaltungsdatei nicht angelegt werden.";
			MyManagementFile.LOGGER.error(errorMessage);
			throw new IOException(errorMessage);
		}
	}

	@Override
	public void close() throws IOException {
		try {
			save();
		} finally {
			// lock-Datei wieder freigeben
			managementFileLock.unlock();
		}
	}

	@Override
	public File getObjectSetDirectory() {
		final File objectSetDirectory = new File(objectSetBaseDirectory, "mengen-" + getConfigurationAuthority());
		if (!objectSetDirectory.isDirectory()) {
			if (!objectSetDirectory.mkdir()) {
				throw new IllegalStateException(
						"Verzeichnis für die Speicherung von dynamischen Mengen konnte nicht erzeugt werden: "
								+ objectSetDirectory.getPath());
			}
		}
		return objectSetDirectory;
	}

	/**
	 * String-Repräsentation (Verzeichnis und Dateiname der Verwaltungsdatei) dieses
	 * Objekts.
	 *
	 * @return String-Repräsentation (Verzeichnis und Dateiname der
	 *         Verwaltungsdatei) dieses Objekts.
	 */
	@Override
	public String toString() {
		return xmlFile.toString();
	}

	/**
	 * Speichert die Verwaltungsdaten im angegebenen Datenstrom ab.
	 *
	 * @param outputStream der Stream, in dem die Verwaltungsdaten abgespeichert
	 *                     werden sollen
	 *
	 * @throws IOException Falls ein Fehler beim Speichern der Verwaltungsdaten
	 *                     aufgetreten ist.
	 */
	private void save(final OutputStream outputStream) throws IOException {
		try {
			final TransformerFactory transformerFactory = TransformerFactory.newInstance();
			final Transformer transformer = transformerFactory.newTransformer();
			transformer.setOutputProperty(OutputKeys.ENCODING, "ISO-8859-1"); // ISO-Kodierung
			// für
			// westeuropäische
			// Sprachen
			transformer.setOutputProperty(OutputKeys.INDENT, "yes");
			transformer.setOutputProperty(OutputKeys.STANDALONE, "no"); // DTD
			// ist
			// in
			// einer
			// separaten
			// Datei

			// DOCTYPE bestimmen
			final DocumentType documentType = xmlDocument.getDoctype();
			String publicID = null;
			String systemID = null;
			if (documentType != null) {
				publicID = xmlDocument.getDoctype().getPublicId();
				systemID = xmlDocument.getDoctype().getSystemId();
			}
			if (publicID != null) {
				transformer.setOutputProperty(OutputKeys.DOCTYPE_PUBLIC, publicID);
			} else {
				// DOCTYPE PUBLIC_ID ist nicht vorhanden -> erstellen
				transformer.setOutputProperty(OutputKeys.DOCTYPE_PUBLIC, "-//K2S//DTD Verwaltung//DE");
			}
			if (systemID != null) {
				transformer.setOutputProperty(OutputKeys.DOCTYPE_SYSTEM, systemID);
			} else {
				// DOCTYPE SYSTEM_ID ist nicht vorhanden -> erstellen
				transformer.setOutputProperty(OutputKeys.DOCTYPE_SYSTEM, "verwaltungsdaten.dtd");
			}

			final DOMSource source = new DOMSource(xmlDocument);
			// StreamResult result = new StreamResult(System.out); // System.out
			// -> gibt die XML-Struktur in einer Shell aus.
			final StreamResult result = new StreamResult(outputStream); // gibt
			// die
			// XML-Struktur
			// in
			// einem
			// Stream
			// (Datei)
			// aus
			transformer.transform(source, result);
			outputStream.close();
		} catch (final Exception ex) {
			final String errorMessage = "Ein Fehler beim Speichern der Verwaltungsdaten ist aufgetreten";
			MyManagementFile.LOGGER.error(errorMessage, ex);
			throw new IOException(errorMessage);
		}
	}

	/**
	 * Diese Implementierung ist für die Konfigurationsbereiche der Verwaltungsdaten
	 * zuständig.
	 */
	private final class MyConfigAreaManagementInfo implements MyConfigurationAreaManagementInfo {

		/**
		 * Die Pid des Konfigurationsbereichs ist unveränderlich und wird hier
		 * gespeichert.
		 */
		private final String pid;

		/** Enthält alle Versionseinträge eines Konfigurationsbereichs. */
		private final List<VersionInfo> versionInfos;

		/** Gibt an, ob eine neue Version aktiviert wurde, oder nicht. */
		private boolean newVersionActivated;

		/** Die aktuelle und aktivierte Version des Konfigurationsbereichs. */
		private VersionInfo versionInfo = new VersionInformation((short) 0, "");

		/**
		 * Konstruktor erstellt einen Eintrag in der XML-Datenstruktur für einen
		 * Konfigurationsbereich.
		 *
		 * @param pid    die Pid des Konfigurationsbereichs
		 * @param create gibt an, ob ein Eintrag in der XML-Datenstruktur hinzugefügt
		 *               werden soll
		 */
		private MyConfigAreaManagementInfo(final String pid, final boolean create) {
			versionInfos = new LinkedList<>();
			this.pid = pid;
			if (create) {
				// neuer Eintrag in xml
				synchronized (xmlDocument) {
					final Element xmlRoot = xmlDocument.getDocumentElement();
					final Element element = xmlDocument.createElement("konfigurationsbereich");
					element.setAttribute("pid", pid);
					final URI uri = uriBase.relativize(xmlFile.getParentFile().toURI());
					element.setAttribute("verzeichnis", uri.toString());
					xmlRoot.appendChild(element);
				}
			}
		}

		/**
		 * Gibt die Pid des Konfigurationsbereichs eines Verwaltungseintrages zurück.
		 *
		 * @return die Pid des Konfigurationsbereichs
		 */
		@Override
		public String getPid() {
			return pid;
		}

		/**
		 * Gibt den Speicherort (Verzeichnis) der Datei des Konfigurationsbereichs
		 * zurück.
		 *
		 * @return der Speicherort (Verzeichnis) der Konfigurationsbereichsdatei
		 */
		@Override
		public File getDirectory() {
			synchronized (xmlDocument) {
				final Element xmlRoot = xmlDocument.getDocumentElement();
				final NodeList configurationAreaList = xmlRoot.getElementsByTagName("konfigurationsbereich");
				for (int i = 0; i < configurationAreaList.getLength(); i++) {
					final Element element = (Element) configurationAreaList.item(i);
					if (element.getAttribute("pid").equals(pid)) {
						// passenden Eintrag gefunden
						try {
							URI uri = new URI(element.getAttribute("verzeichnis"));
							if (!uri.isAbsolute()) {
								uri = uriBase.resolve(uri);
							}
							final File file = new File(uri);
							if (file.isDirectory()) {
								return file;
							}
							throw new IllegalStateException(
									"Das ausgelesene Verzeichnis ist keines oder existiert nicht: " + file.toString());
						} catch (final URISyntaxException ex) {
							final String errorMessage = "Speicherort des Konfigurationsbereichs '" + pid
									+ "' kann nicht ermittelt werden";
							MyManagementFile.LOGGER.warning(errorMessage, ex);
							throw new IllegalStateException(errorMessage, ex);
						}
					}
				}
			}
			throw new IllegalStateException("Der Konfigurationsbereich mit dieser Pid '" + pid
					+ "' ist nicht in den Verwaltungsdaten vorhanden.");
		}

		/**
		 * Speichert den Speicherort (Verzeichnis) der Datei des Konfigurationsbereichs.
		 * Stimmt ein Teil des angegebenen Verzeichnisses mit dem vollständigen
		 * Verzeichnis der Verwaltungsdatei überein, so wird die Pfadangabe relativ zum
		 * Verzeichnis der Verwaltungsdatei abgespeichert.
		 *
		 * @param directory der Speicherort (Verzeichnis) der
		 *                  Konfigurationsbereichsdatei
		 */
		@Override
		public void setDirectory(final File directory) {
			synchronized (xmlDocument) {
				final Element xmlRoot = xmlDocument.getDocumentElement();
				final NodeList configurationAreaList = xmlRoot.getElementsByTagName("konfigurationsbereich");
				for (int i = 0; i < configurationAreaList.getLength(); i++) {
					final Element element = (Element) configurationAreaList.item(i);
					if (element.getAttribute("pid").equals(pid)) {
						// passenden Eintrag gefunden
						if (directory.isDirectory()) {
							final URI uri = uriBase.relativize(directory.toURI());
							element.setAttribute("verzeichnis", uri.toString());
						} else {
							throw new IllegalArgumentException(
									"Der angegebene Speicherort ist kein Verzeichnis oder existiert nicht: "
											+ directory.toString());
						}
						break;
					}
				}
			}
		}

		@Override
		public void setNextActiveVersion(final short nextActiveVersion) {
			if (nextActiveVersion < 1) {
				throw new IllegalArgumentException("Die nächste zu aktivierende Version muss größer gleich 1 sein.");
			}
			synchronized (xmlDocument) {
				final Element xmlRoot = xmlDocument.getDocumentElement();
				final NodeList configList = xmlRoot.getElementsByTagName("konfigurationsbereich");
				for (int i = 0; i < configList.getLength(); i++) {
					final Element element = (Element) configList.item(i);
					if (element.getAttribute("pid").equals(pid)) {
						// richtigen Konfigurationsbereich gefunden
						final NodeList versionList = element.getElementsByTagName("version");
						// prüfen, ob die Versionsnummer größer ist, als die
						// bisher aktivierte
						if (versionList.getLength() > 0) {
							// es gibt mindestens einen Versionseintrag
							final Element version = (Element) versionList.item(versionList.getLength() - 1);
							final short nr = Short.parseShort(version.getAttribute("nr"));
							final String time = version.getAttribute("zeitpunkt");
							// wurde die letzte Versionsnummer bereits
							// aktiviert?
							if (time.isEmpty()) {
								// letzte Versionsnummer wurde noch nicht
								// aktiviert
								// gibt es einen Wert davor?
								if (versionList.getLength() > 1) {
									final Element preVersion = (Element) versionList.item(versionList.getLength() - 2);
									final short preNr = Short.parseShort(preVersion.getAttribute("nr"));
									if (nextActiveVersion > preNr) {
										// ändere die geschriebene
										// Versionsnummer
										version.setAttribute("nr", String.valueOf(nextActiveVersion));
									} else {
										// Exception
										final StringBuilder errorMessage = new StringBuilder();
										errorMessage.append("Die nächste zu aktivierende Version ")
												.append(nextActiveVersion);
										errorMessage.append(" ist kleiner oder gleich der aktiven Version ")
												.append(preNr);
										MyManagementFile.LOGGER.error(errorMessage.toString());
										throw new IllegalArgumentException(errorMessage.toString());
									}
								} else {
									version.setAttribute("nr", String.valueOf(nextActiveVersion));
								}
							} else {
								// letzte Versionsnummer wurde bereits aktiviert
								// -> neuen Eintrag erstellen
								if (nextActiveVersion > nr) {
									// neuen Eintrag erstellen
									createNextVersion(element, nextActiveVersion);
								} else {
									// Exception
									final StringBuilder errorMessage = new StringBuilder();
									errorMessage.append("Die nächste zu aktivierende Version ")
											.append(nextActiveVersion);
									errorMessage.append(" ist kleiner oder gleich der aktiven Version ").append(nr);
									MyManagementFile.LOGGER.error(errorMessage.toString());
									throw new IllegalArgumentException(errorMessage.toString());
								}
							}
						} else {
							// es gibt noch keinen Versionseintrag -> neuen
							// Eintrag erstellen
							createNextVersion(element, nextActiveVersion);
						}
					}
				}
			}
		}

		@Override
		public short delAllVersionInfo() {
			synchronized (xmlDocument) {
				short nr = 0;
				final Element xmlRoot = xmlDocument.getDocumentElement();
				final NodeList configList = xmlRoot.getElementsByTagName("konfigurationsbereich");
				for (int i = 0; i < configList.getLength(); i++) {
					final Element element = (Element) configList.item(i);
					if (element.getAttribute("pid").equals(pid)) {
						final NodeList versionList = element.getElementsByTagName("version");

						if (versionList.getLength() > 0) {
							final Element version = (Element) versionList.item(versionList.getLength() - 1);
							nr = Short.parseShort(version.getAttribute("nr"));

							final List<Element> delList = new LinkedList<>();
							for (int loc = 0; loc < versionList.getLength(); loc++) {
								final Element element2 = (Element) versionList.item(loc);
								delList.add(element2);
							}

							for (final Element element2 : delList) {
								element.removeChild(element2);
							}
							createNextVersion(element, nr);
							final NodeList newVersionList = element.getElementsByTagName("version");
							if (newVersionList.getLength() > 0) {
								final Element newElement = (Element) versionList.item(versionList.getLength() - 1);
								newElement.setAttribute("zeitpunkt", "");
							}
						}
					}
				}
				return nr;
			}
		}

		/**
		 * Erstellt einen neuen Versionseintrag in der Verwaltungsdatei.
		 *
		 * @param node    KonfigurationsBereichs-Knoten, an dem der Versionseintrag
		 *                gespeichert werden soll
		 * @param version zu speichernde Versionsnummer
		 */
		private void createNextVersion(final Node node, final short version) {
			// neuen Versionsknoten anlegen
			final Element nextVersion = xmlDocument.createElement("version");
			nextVersion.setAttribute("nr", String.valueOf(version));

			// Knoten hinzufügen
			node.appendChild(nextVersion);
		}

		@Override
		public boolean isNewVersionActivated() {
			return newVersionActivated;
		}

		/**
		 * Setzt den Parameter, ob dieser Konfigurationsbereich in eine neue Version
		 * überführt wurde.
		 *
		 * @param newVersionActivated ob dieser Konfigurationsbereich in eine neue
		 *                            Version überführt wurde
		 */
		private void setNewVersionActivated(final boolean newVersionActivated) {
			this.newVersionActivated = newVersionActivated;
		}

		/**
		 * Gibt die aktive Version des Konfigurationsbereichs und ihren
		 * Aktivierungszeitpunkt zurück.
		 *
		 * @return die aktive Version und ihren Aktivierungszeitpunkt des
		 *         Konfigurationsbereichs
		 */
		@Override
		public VersionInfo getActiveVersion() {
			return versionInfo;
		}

		/**
		 * Interne Methode, die die aktive Version des Konfigurationsbereichs am
		 * Verwaltungseintrag speichert.
		 *
		 * @param info der Versionseintrag mit der aktiven Version
		 */
		private void setActiveVersion(final VersionInfo info) {
			this.versionInfo = info;
		}

		/**
		 * Gibt alle Versionseinträge zu diesem Konfigurationsbereich in einer Liste
		 * zurück.
		 *
		 * @return eine Liste aller Versionseinträge zu diesem Konfigurationsbereich
		 */
		@Override
		public List<VersionInfo> getVersions() {
			return versionInfos;
		}

		/**
		 * Gibt die Position innerhalb aller Konfigurationsbereiche in den
		 * Verwaltungsdaten zurück. (siehe auch TPuK1-99 Reihenfolge der Bereiche)
		 *
		 * @return Position innerhalb der Konfigurationsbereiche
		 */
		@Override
		public int getPosition() {
			synchronized (xmlDocument) {
				final Element xmlRoot = xmlDocument.getDocumentElement();
				final NodeList nodeList = xmlRoot.getElementsByTagName("konfigurationsbereich");
				for (int i = 0; i < nodeList.getLength(); i++) {
					final Element element = (Element) nodeList.item(i);
					if (element.getAttribute("pid").equals(pid)) {
						return i + 1;
					}
				}
			}
			throw new IllegalStateException("Der Konfigurationsbereich mit dieser Pid '" + pid
					+ "' ist nicht in den Verwaltungsdaten vorhanden.");
		}

		/**
		 * Setzt die Position innerhalb der Reihenfolge der Konfigurationsbereiche.
		 * (siehe auch TPuK1-99 Reihenfolge der Bereiche)
		 *
		 * @param position Position innerhalb der Konfigurationsbereiche
		 */
		@Override
		public void setPosition(final int position) {
			if (position <= 0) {
				throw new IllegalArgumentException("Ungültige Position! Die erste Position ist die 1.");
			}
			// Eintrag in der xml-Datei verschieben! -> passenden Eintrag finden
			// und woanders hin kopieren
			synchronized (xmlDocument) {
				final Element xmlRoot = xmlDocument.getDocumentElement();
				// erst das Element rauslöschen
				final NodeList configurationAreaList = xmlRoot.getElementsByTagName("konfigurationsbereich");
				if (position > configurationAreaList.getLength()) {
					throw new IllegalArgumentException(
							"Ungültige Position! Es gibt nur " + configurationAreaList.getLength() + " Einträge.");
				}

				Node nodeToMove = null;
				for (int i = 0; i < configurationAreaList.getLength(); i++) {
					final Element element = (Element) configurationAreaList.item(i);
					if (element.getAttribute("pid").equals(pid)) {
						nodeToMove = configurationAreaList.item(i);
						break;
					}
				}
				if (nodeToMove != null) {
					xmlRoot.removeChild(nodeToMove);
					// dann das Element einfügen
					final NodeList newList = xmlRoot.getElementsByTagName("konfigurationsbereich");
					xmlRoot.insertBefore(nodeToMove, newList.item(position - 1));
				}
			}
		}

		/**
		 * Interne Methode, die die Versionseinträge einliest und in einer Liste
		 * speichert.
		 *
		 * @param info ein Versionseintrag
		 */
		private void addVersionInfo(final VersionInfo info) {
			versionInfos.add(info);
		}
	}

	/**
	 * Repräsentiert einen Versionseintrag in den Verwaltungsdaten zu einem
	 * Konfigurationsbereich.
	 */
	private final class VersionInformation implements VersionInfo {

		/** Versionsnummer */
		private final short version;

		/** Aktivierungszeit dieses Versionseintrages */
		private String activationTime = "";

		private VersionInformation(final short version, final String activationTime) {
			this.version = version;
			this.activationTime = activationTime;
		}

		/**
		 * Gibt die Versionsnummer dieses Versionseintrages zurück.
		 *
		 * @return die Versionsnummer
		 */
		@Override
		public short getVersion() {
			return version;
		}

		/**
		 * Gibt den Zeitpunkt der Aktivierung dieser Version zurück.
		 *
		 * @return den Aktivierungszeitpunkt
		 */
		@Override
		public long getActivationTime() {
			try {
				if (dateFormat == null) {
					dateFormat = new SimpleDateFormat("dd.MM.yyyy HH:mm:ss,SSS");
				}
				final Date date = dateFormat.parse(activationTime);
				return date.getTime();
			} catch (final ParseException ex) {
				final String errorMessage = "Die Aktivierungszeit " + activationTime
						+ " konnte nicht eingelesen werden. Folgendes Format verwenden 'dd.MM.yyyy HH:mm:ss,SSS'.";
				MyManagementFile.LOGGER.error(errorMessage);
				throw new IllegalStateException(errorMessage, ex);
			}
		}
	}

	/**
	 * Implementierung eines EntityResolvers, der Referenzen auf den
	 * Public-Identifier "-//K2S//DTD Verwaltung//DE" ersetzt durch die
	 * verwaltungsdaten.dtd Resource-Datei in diesem Package.
	 */
	private static class ManagementEntityResolver implements EntityResolver {

		/**
		 * Löst Referenzen auf external entities wie z.B. DTD-Dateien auf.
		 * <p/>
		 * Angegebene Dateien werden, falls sie im Suchverzeichnis gefunden werden, von
		 * dort geladen. Ansonsten wird der normale Mechanismus zum Laden von externen
		 * Entities benutzt.
		 *
		 * @param publicId Der public identifier der externen Entity oder null falls
		 *                 dieser nicht verfügbar ist.
		 * @param systemId Der system identifier aus dem XML-Dokument.
		 *
		 * @return Für Referenzen im Suchverzeichnis wird ein InputSource-Objekt, das
		 *         mit der entsprechenden Datei im Suchverzeichnis verbunden ist,
		 *         zurückgegeben.
		 *
		 * @throws SAXException Bei Fehlern beim Zugriff auf externe Entities.
		 * @throws IOException
		 */
		@Override
		public InputSource resolveEntity(final String publicId, final String systemId)
				throws SAXException, IOException {

			if ("-//K2S//DTD Verwaltung//DE".equals(publicId)) {
				final URL url = this.getClass().getResource("verwaltungsdaten.dtd");
				assert url != null : this.getClass();
				return new InputSource(url.toExternalForm());
			}

			return null;
		}
	}

	@Override
	public File getForeignObjectCacheFile() {
		return null;
	}

	@Override
	public List<ConfigurationAreaManagementInfo> getAllConfigurationAreaManagementInfos() {
		return new ArrayList<>(areaManagementInfos);
	}
}
